import {relative, dirname} from 'path';
import PrintOptions, {
  isDefaultExportCandidate,
  resolveExportName,
  resolveFilename,
} from './PrintOptions';
import FileName from './FileName';
import IdentifierName from './IdentifierName';

export interface FileContext {
  // asExport: (declaration: string[]) => string[];
  // asNamedExport: (declaration: string[]) => string[];
  // asDefaultExport: (exportName: string) => string[];
  getImport: (fileExport: FileExport) => string;
}
export interface FileExport {
  mode: 'type' | 'value';
  file: FileName;
  isDefaultExport: () => boolean;
  exportName: string;
}

class ImportState {
  public readonly file: FileName;
  private readonly _namedImports = new Map<string, FileExport>();
  constructor(file: FileName) {
    this.file = file;
  }
  public getImport(fileExport: FileExport): string {
    this._namedImports.set(fileExport.exportName, fileExport);
    return fileExport.exportName;
  }
  public getImportStatement(relativePath: string) {
    const defaultImport = [...this._namedImports.values()].find((v) =>
      v.isDefaultExport(),
    )?.exportName;
    const namedImports = [...this._namedImports.values()]
      .filter((v) => !v.isDefaultExport())
      .map((v) => v.exportName);
    const specifiers: string[] = [];
    if (defaultImport) {
      specifiers.push(defaultImport);
    }
    if (namedImports.length) {
      specifiers.push(`{${namedImports.sort().join(', ')}}`);
    }
    return `import ${specifiers.join(', ')} from '${relativePath}'`;
  }
}

class FileContent<TypeID> {
  public readonly file: FileName;
  private readonly _options: PrintOptions<TypeID>;
  private readonly _imports = new Map<FileName, ImportState>();
  private readonly _declarationNames = new Set<string>();
  private readonly _declarations: {name: string; src: () => string[]}[] = [];
  private readonly _reExports: {
    type: {source: string; dest: string}[];
    value: {source: string; dest: string}[];
  } = {type: [], value: []};
  constructor(file: FileName, options: PrintOptions<TypeID>) {
    this.file = file;
    this._options = options;
  }
  private _defaultCandidate: TypeID | undefined;
  private _defaultName: IdentifierName | null | undefined;
  private getDefaultName(): IdentifierName | null {
    if (this._defaultName === undefined) {
      this._defaultName = this._defaultCandidate
        ? resolveExportName(this._defaultCandidate, this._options)
        : null;
    }
    return this._defaultName;
  }

  private _getImportState(file: FileName): {
    getImport: (fileExport: FileExport) => string;
  } {
    if (file === this.file) {
      return {
        getImport: (n) => n.exportName,
      };
    }

    const cachedImportState = this._imports.get(file);
    if (cachedImportState !== undefined) {
      return cachedImportState;
    }

    const newImportState = new ImportState(file);
    this._imports.set(file, newImportState);
    return newImportState;
  }

  public pushDeclaration(
    typeID: TypeID,
    mode: 'type' | 'value',
    declaration: (identifier: IdentifierName, imp: FileContext) => string[],
  ): FileExport {
    const identifierName = resolveExportName(typeID, this._options);
    if (!this._declarationNames.has(identifierName)) {
      if (
        this._defaultName === undefined &&
        isDefaultExportCandidate(typeID, this._options)
      ) {
        if (this._defaultCandidate === undefined) {
          this._defaultCandidate = typeID;
        } else {
          const oldWeight = this._options.getExportPriority(
            this._defaultCandidate,
          );
          const newWeight = this._options.getExportPriority(typeID);
          if (oldWeight === newWeight) {
            this._defaultName = null;
          } else if (oldWeight > newWeight) {
            this._defaultCandidate = typeID;
          }
        }
      }
      this._declarationNames.add(identifierName);
      const declarationLines = declaration(identifierName, {
        getImport: (id: FileExport) =>
          this._getImportState(id.file).getImport(id),
      });
      this._declarations.push({
        name: identifierName,
        src: () => [
          ...declarationLines,
          this.getDefaultName() === identifierName
            ? `export default ${identifierName};`
            : `export ${mode === 'type' ? 'type ' : ''}{${identifierName}}`,
        ],
      });
    }
    return {
      mode,
      file: this.file,
      isDefaultExport: () => this.getDefaultName() === identifierName,
      exportName: identifierName,
    };
  }
  public pushReExport(dest: TypeID, source: FileExport) {
    const identifierName = resolveExportName(dest, this._options);
    if (this._declarationNames.has(identifierName)) {
      return;
    }
    const importedName = this._getImportState(source.file).getImport(source);
    this._declarationNames.add(identifierName);
    this._reExports[source.mode].push({
      source: importedName,
      dest: identifierName,
    });
  }

  public getContent() {
    return (
      [
        ...(this._imports.size
          ? [
              [...this._imports.values()]
                .sort((a, b) => (a.file < b.file ? -1 : 1))
                .map((imp) => {
                  const relativePath = relative(dirname(this.file), imp.file);
                  return imp.getImportStatement(
                    `${
                      relativePath[0] === '.' ? '' : './'
                    }${relativePath.replace(/(\.d)?\.tsx?$/, '')}`,
                  );
                })
                .join('\n'),
            ]
          : []),
        ...this._declarations
          .sort((a, b) => (a.name < b.name ? -1 : 1))
          .map((v) => v.src().join('\n')),
        ...(this._reExports.type.length
          ? [
              `export type {\n${this._reExports.type
                .sort((a, b) => (a.dest < b.dest ? -1 : 1))
                .map((t) =>
                  t.source === t.dest
                    ? `  ${t.source},`
                    : `  ${t.source} as ${t.dest},`,
                )
                .join(`\n`)}\n}`,
            ]
          : []),
        ...(this._reExports.value.length
          ? [
              `export {\n${this._reExports.value
                .sort((a, b) => (a.dest < b.dest ? -1 : 1))
                .map((t) =>
                  t.source === t.dest
                    ? `  ${t.source},`
                    : `  ${t.source} as ${t.dest},`,
                )
                .join(`\n`)}\n}`,
            ]
          : []),
      ].join('\n\n') + '\n'
    );
  }
}

export default class PrintContext<TypeID> {
  private readonly _files = new Map<FileName, FileContent<TypeID>>();
  private readonly _rawFiles = new Map<FileName, string>();

  public readonly options: PrintOptions<TypeID>;
  constructor(options: PrintOptions<TypeID>) {
    this.options = options;
  }

  private _pushDeclaration(
    id: TypeID,
    mode: 'type' | 'value',
    declaration: (identifier: IdentifierName, imp: FileContext) => string[],
  ): FileExport {
    const file = resolveFilename(id, this.options);
    const fileContent = mapGetOrSet(
      this._files,
      file,
      () => new FileContent(file, this.options),
    );
    return fileContent.pushDeclaration(id, mode, declaration);
  }

  public pushTypeDeclaration(
    id: TypeID,
    declaration: (identifier: IdentifierName, imp: FileContext) => string[],
  ): FileExport {
    return this._pushDeclaration(id, 'type', declaration);
  }
  public pushReExport(id: TypeID, from: FileExport): void {
    const file = resolveFilename(id, this.options);
    const fileContent = mapGetOrSet(
      this._files,
      file,
      () => new FileContent(file, this.options),
    );
    fileContent.pushReExport(id, from);
  }

  public pushValueDeclaration(
    id: TypeID,
    declaration: (identifier: IdentifierName, imp: FileContext) => string[],
  ): FileExport {
    return this._pushDeclaration(id, 'value', declaration);
  }

  public writeFile(filename: FileName, content: string) {
    if (this._rawFiles.has(filename)) {
      throw new Error(`Cannot write the same file multiple times: ${filename}`);
    }
    this._rawFiles.set(filename, content);
  }
  public getFiles() {
    return [
      ...[...this._files.values()].map((file) => ({
        filename: file.file,
        content: file.getContent(),
      })),
      ...[...this._rawFiles].map(([filename, content]) => ({
        filename,
        content,
      })),
    ];
  }
}

function mapSetAndReturn<TKey, TValue>(
  map: {
    set(key: TKey, value: TValue): unknown;
  },
  key: TKey,
  value: TValue,
) {
  map.set(key, value);
  return value;
}

function mapGetOrSet<TKey, TValue>(
  map: {
    get(key: TKey): TValue | undefined;
    set(key: TKey, value: TValue): unknown;
  },
  key: TKey,
  value: () => TValue,
) {
  const cached = map.get(key);
  if (cached !== undefined) return cached;
  return mapSetAndReturn(map, key, value());
}
